import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

//Mathieu Gascon 5/13/2021
//Input : The first input is a file containing gene trees in newick and Ensembl Multi Format (.nh.emf) and the second input is a file containing a species tree in newick format (with all but root nodes having a distance to parent)
//Output : Statistics on the pruned trees obtained using the MULPT algorithm (greedy) and an exact algorithm

//Note : This implementation of the greedy algorithm (MULPR) is in O(n^2 * d^2) where n is the size of the gene tree and d the number of duplication nodes in the reconciliation with the species tree.
public class Main {

	public static void main(String[] args) throws IOException{
		
		//Initializing statistics
		int numDupGreedy = 0;
		int numDupExact = 0;
		int numDupAfterPreprocess = 0;
		int numTree = 0;
		int numTreeWithDupAfterPreProcess = 0;
		int numTreePreProcessExact = 0;
		ArrayList<ArrayList<Double>> ratio = new ArrayList<ArrayList<Double>>();
		for(int i =0; i<100; i++) {
           ratio.add(new ArrayList<Double>());
		}
		int[] difference = new int[100];
		int[] differenceNotExact = new int[100];
		long greedyTime = 0;
		long exactTime = 0;
		int k = 0;
		
		String speciesTreeFileName = args[1];
		String geneTreesFileName = args[0];
		
		//Get the number of gene trees of the file given in input
		int numGeneTrees = getNumGeneTrees(geneTreesFileName);
			
		//Get the species tree givent in input
		String fileName2 = speciesTreeFileName;
        File file2 = new File(fileName2);	
		BufferedReader Buff2 = new BufferedReader(new FileReader(file2));      
		String speciesTree1 = Buff2.readLine();
		String nouvelspeciesTree = "";
		for(int i = 0; i < speciesTree1.length();i++) {
			String currentCaracter = speciesTree1.substring(i, i+1);
			if(!currentCaracter.equals(":")) {
			nouvelspeciesTree += currentCaracter;
			} else {
				for(int j = i+1; j< speciesTree1.length(); j++ ) {
					i++;
					if(speciesTree1.substring(j, j+1).equals(",") || speciesTree1.substring(j, j+1).equals(")")) {
						nouvelspeciesTree += speciesTree1.substring(j, j+1);
						break;
					}
				}
			}
		
		
		}
		speciesTree1 = nouvelspeciesTree;	
		Tree speciesTree = creerspeciesTree(speciesTree1);
		chooseSpeciesNodeInternalSpecies(speciesTree);
		findCladesSpecies(speciesTree);
		findParent(speciesTree);
		findLeavesToPrune(speciesTree);
		PreProcessS(speciesTree);
		HashMap<String, Integer> speciesNoList = new HashMap<String, Integer>();
		HashMap<String, Node> speciesNodesList = new HashMap<String, Node>();
		computeSpeciesNoList(speciesTree.getRoot(), speciesNoList);
		computeSpeciesNodesList(speciesTree.getRoot(), speciesNodesList);
		
		//Open the gene trees file
		String fileName = geneTreesFileName;
        File file = new File(fileName);	
		BufferedReader Buff = new BufferedReader(new FileReader(file));      
		
		for(k = 0; k < numGeneTrees ;k++) {
			//System.out.println(k);
			
			//Get the current gene tree
			String st;
			String geneTree1 = "";
			HashMap<String, String> listauSpecies = new HashMap<String, String>();
			while (((st = Buff.readLine())!= null)){
				if(st.equals("DATA")) {
					st = Buff.readLine();
					geneTree1 = st;
					break;
				}
				
				if((st.length() >= 3) && (st.substring(0, 3).equals("SEQ"))) {
				String[] array = st.split(" "); 
				listauSpecies.put(array[2], array[1]);
				}
			}
			
			String newGeneTree = "";
			for(int i = 0; i < geneTree1.length();i++) {
				String currentCaracter = geneTree1.substring(i, i+1);
				if(!currentCaracter.equals(":")) {
				newGeneTree += currentCaracter;
				} else {
					for(int j = i+1; j< geneTree1.length(); j++ ) {
						i++;
						if(geneTree1.substring(j, j+1).equals(",") || geneTree1.substring(j, j+1).equals(")")) {
							newGeneTree += geneTree1.substring(j, j+1);
							break;
						}
					}
				}
			
			
			}
			geneTree1 = newGeneTree;
			//System.out.println(geneTree1);
			newGeneTree = "";
			for(int i = 0; i < geneTree1.length();i++) {
				String currentCaracter = geneTree1.substring(i, i+1);
				if(currentCaracter.equals(",") || currentCaracter.equals("(") || currentCaracter.equals(")") ) {
				newGeneTree += currentCaracter;
				} else {
					String Species = currentCaracter;
					for(int j = i+1; j< geneTree1.length(); j++ ) {
						i++;
						if(!geneTree1.substring(j, j+1).equals(",") && !geneTree1.substring(j, j+1).equals(")") && !geneTree1.substring(j, j+1).equals("(")) {
						Species += geneTree1.substring(j, j+1);
						}
						if(geneTree1.substring(j, j+1).equals(",") || geneTree1.substring(j, j+1).equals(")") || geneTree1.substring(j, j+1).equals("(")) {
							newGeneTree += listauSpecies.get(Species);
							newGeneTree += geneTree1.substring(j, j+1);
							break;
						}
					}
			}
			}
			geneTree1 = newGeneTree + ";";
			
			Tree geneTree = creergeneTree(
				geneTree1,
					geneTree1);
			findCladesSpecies(geneTree);
			chooseSpeciesNodeInternalGene(geneTree,speciesTree);
			findParent(geneTree);
			findNoNodeGene(geneTree);
			
			
			//Filtrer l'tree de gene pour ne garder que les feuilles d'labels se trouvant dans l'tree d'Species.
			String speciesTreeString = speciesTree.getRoot().toString3();
			speciesTreeString = speciesTreeString.replaceAll("\\)", "");
			speciesTreeString = speciesTreeString.replaceAll("\\(", "");
			speciesTreeString = speciesTreeString.replaceAll(";", "");
			String[] listSpecies = speciesTreeString.split(","); 		
			
			Set<String> HashSpecies = new HashSet<String>();
			for(int j = 0; j < listSpecies.length; j++) {
				HashSpecies.add(listSpecies[j]);
			}
			
		//	System.out.println("Number of species : " + HashSpecies.size());
	//		System.out.println();
			
			try {
			FiltrergeneTree(geneTree,HashSpecies);
			} catch (Exception e) {
			//	System.out.println("gene tree empty");
		//		System.out.println();
				continue;
			}
			
			findCladesSpecies(geneTree);
			chooseSpeciesNodeInternalGene(geneTree,speciesTree);
			findParent(geneTree);
			findDuplicationNodes(geneTree,speciesNoList,speciesNodesList);
			findNoNodeGene(geneTree);
			
			// To print the pruned trees before the preprocess, uncomment the lines below	
			//	System.out.println("Before preProcess: ");
			//	System.out.println(geneTree);
			//	System.out.println(geneTree.getNumDuplications());
			//System.out.println();
			
			//Preprocessing of the current gene tree 
			preProcessGeneTree(geneTree,speciesTree,speciesNoList,speciesNodesList);

		// To print the pruned trees after the preprocess, uncomment the lines below	
		//	System.out.println("After preProcess: ");
		//	System.out.println(geneTree);
		//	System.out.println(geneTree.getNumDuplications());
		//	System.out.println();
			
			int preProcessDup = geneTree.getNumDuplications();
			
			if(preProcessDup != 0) {
				numTreeWithDupAfterPreProcess++;
				numDupAfterPreprocess = numDupAfterPreprocess+preProcessDup;
			}
					
			if(geneTree.getRoot().getRightChild() != null) {
				numTree++;
				
				findCladesSpecies(geneTree);
				chooseSpeciesNodeInternalGene(geneTree,speciesTree);
				findParent(geneTree);
				findDuplicationNodes(geneTree,speciesNoList,speciesNodesList);
				findNoNodeGene(geneTree);
			
				
				long start = System.currentTimeMillis();
			Tree glouton = pruningGreedy(geneTree,speciesTree,speciesNoList,speciesNodesList);
			greedyTime = greedyTime + (System.currentTimeMillis()-start);
		// To print the pruned trees after the greedy algorithm, uncomment the lines below
		//	System.out.println("After pruningGreedy (Greedy): ");
		//	System.out.println(glouton);
		//	System.out.println(glouton.getNumDuplications());
		//System.out.println();
			numDupGreedy = numDupGreedy + glouton.getNumDuplications();
			
			
		
			
			long start2 = System.currentTimeMillis();
			Tree exacte = pruningExact(geneTree,speciesTree,speciesNoList,speciesNodesList);
			exactTime = exactTime + (System.currentTimeMillis()-start2);
		//  To print the pruned trees after the exact algorithm, uncomment the lines below
		//	System.out.println("After pruningGreedy (Exact): ");	
		//	System.out.println(exacte);
		//	System.out.println(exacte.getNumDuplications());
			numDupExact = numDupExact + exacte.getNumDuplications();
			
			
			
			if(preProcessDup == exacte.getNumDuplications()) {
				numTreePreProcessExact++;
			} else {
				differenceNotExact[glouton.getNumDuplications()-exacte.getNumDuplications()]++;
			}
			
			difference[glouton.getNumDuplications()-exacte.getNumDuplications()]++;
			
			if(glouton.getNumDuplications() != 0) {
				ratio.get(preProcessDup).add((double)exacte.getNumDuplications()/(double)glouton.getNumDuplications());
			} else {
				ratio.get(preProcessDup).add(1.0);
			}
			} else {
				
			}
		}


		
		System.out.println("Number of duplications after Greedy algorithm = " + numDupGreedy);
		System.out.println("Number of duplications after exact algorithm = " + numDupExact);
		System.out.println("Number of duplications after preprocess algorithm = " + numDupAfterPreprocess);
		System.out.println("Number of non empty trees after preprocess = " + numTree);
		System.out.println("Number of trees with at least one duplication after preprocess = " + numTreeWithDupAfterPreProcess);
		System.out.println("Number of trees for which the preprocess returned the exact solutions = " + numTreePreProcessExact);
		System.out.println("Total execution time of the greedy algorithm (in ms) = " + greedyTime);
		System.out.println("Total execution time of the exact algorithm (in ms) = " + exactTime);
		System.out.println();
		System.out.println("Difference between Greedy and exact algorithm for all trees (" + numTree + " trees)");
		System.out.println();
		for(int i = 0; i< 30; i++) {
			System.out.println("Number of trees with a difference of " + i + " duplication nodes : " + difference[i]);
		}
		System.out.println();
		System.out.println("Difference between Greedy and exact algorithm for trees whose preprocess did not returned the optimal solution (" + (numTree-numTreePreProcessExact) + " trees)");
		System.out.println();
		for(int i = 0; i< 30; i++) {
			System.out.println("Number of trees with a difference of " + i + " duplication nodes : " + differenceNotExact[i]);
		}
		
		System.out.println();
		System.out.println("Mean ratio of numDupExact/numDupGreedy to the number of duplication nodes after preprocessing (-1 if no tree with that number of duplications after preprocessing)");
		System.out.println();
		for(int i = 0; i< 30; i++) {
			ArrayList<Double> courant = ratio.get(i);
			if(courant.size() == 0) {
				System.out.println("ratio for " + i +  " duplication nodes : " + -1);
			} else {
				double r = 0;
				for(int j = 0; j < courant.size(); j++) {
					r = r + courant.get(j);
				}
				r = r/(double)courant.size();
				System.out.println("ratio for " + i +  " duplication nodes : " + r);
			}
		}
		Buff.close();
		Buff2.close();
	}
	
	//Return the number of trees in the file given in input
	public static int getNumGeneTrees(String fileName) throws IOException {
        File file = new File(fileName);	
		BufferedReader Buff = new BufferedReader(new FileReader(file));      
		String st;
		int NumGeneTrees = 0;
		while (((st = Buff.readLine())!= null)){
			if(st.equals("DATA")) {
				NumGeneTrees++;
			}
		}
		Buff.close();
		return NumGeneTrees;
	}
		
	//Preprocess the gene tree to remove duplications that obviously lead to an optimal solution
	public static void preProcessGeneTree(Tree geneTree, Tree speciesTree,HashMap<String, Integer> speciesNoList,HashMap<String, Node> speciesNodesList) {
		Node root = geneTree.getRoot();
		preProcessGeneTreeRecursive(root,geneTree, speciesTree,speciesNoList,speciesNodesList);
		findCladesSpecies(geneTree);
	//	findCladeslabel(geneTree);
		chooseSpeciesNodeInternalGene(geneTree,speciesTree);
		findParent(geneTree);
		findDuplicationNodes(geneTree,speciesNoList,speciesNodesList);
		
		findNoNodeGene(geneTree);
			
	}
	
	//Return the number of duplications under the node given in input
	public static int numDup(Node node) {
		Integer[] numDup = new Integer[1];
		numDup[0] = 0;
		numDupRecursive(node,numDup);
		return numDup[0];
	}
	
	public static void numDupRecursive(Node node, Integer[] numDup) {
		if(node.getRightChild() != null) {
			numDupRecursive(node.getLeftChild(), numDup);
			numDupRecursive(node.getRightChild(), numDup);
			if(node.isDuplication()) {
				numDup[0] = numDup[0] + 1;
			}
		}
	}
	
	public static void preProcessGeneTreeRecursive(Node currentNode,Tree geneTree, Tree speciesTree,HashMap<String, Integer> speciesNoList,HashMap<String, Node> speciesNodesList) {
		if(currentNode.getRightChild() != null) {
			preProcessGeneTreeRecursive(currentNode.getLeftChild(),geneTree, speciesTree,speciesNoList,speciesNodesList);
			preProcessGeneTreeRecursive(currentNode.getRightChild(),geneTree, speciesTree,speciesNoList,speciesNodesList);	
			if(currentNode.isDuplication()) {
				int numDupLeft = numDup(currentNode.getLeftChild());
				int numDupRight= numDup(currentNode.getRightChild());
				if((numDupLeft == 0) && (numDupRight== 0)) { 
					findClade (currentNode.getLeftChild());
					findClade (currentNode.getRightChild());
					Set<String> cladelabel1 =  currentNode.getLeftChild().getCladeLabel();
					Set<String> cladelabel2 =  currentNode.getRightChild().getCladeLabel();
					if(cladelabel1.containsAll(cladelabel2)) {
						Node parent = currentNode.getParent();
						Node child = currentNode.getLeftChild();
						
						if(parent != null) {
							if(parent.getRightChild() == currentNode) {
								parent.setRightChild(child);
							} else {
								parent.setLeftChild(child);
							}
						} else {
							geneTree.setRoot(child);
						}
						
						findCladesSpecies(geneTree);
						chooseSpeciesNodeInternalGene(geneTree,speciesTree);
						findParent(geneTree);
						findDuplicationNodes(geneTree,speciesNoList,speciesNodesList);
						findNoNodeGene(geneTree);
					} else if (cladelabel2.containsAll(cladelabel1)) {
						Node parent = currentNode.getParent();
						Node child = currentNode.getRightChild();
						
						if(parent != null) {		
							if(parent.getRightChild() == currentNode) {
								parent.setRightChild(child);
							} else {
								parent.setLeftChild(child);
							}
						} else {
							geneTree.setRoot(child);
						}		
						findCladesSpecies(geneTree);
						chooseSpeciesNodeInternalGene(geneTree,speciesTree);
						findParent(geneTree);
						findDuplicationNodes(geneTree,speciesNoList,speciesNodesList);
						findNoNodeGene(geneTree);
					}
				}
			}
		}		
	}
	
	//Exact algorithm that prune the leaves in geneTree in a way that fixes the most duplication nodes.
	public static Tree pruningExact(Tree geneTree, Tree speciesTree,HashMap<String, Integer> speciesNoList,HashMap<String, Node> speciesNodesList) {
		Map<String,ArrayList<Node>> listLabels = new HashMap<String,ArrayList<Node>>(); 
		findLeavesRecursive(geneTree.getRoot(),listLabels);
		Tree minDup = new Tree(null);
		minDup.setNumDuplications(2147483647);
		ArrayList<Tree> min = new ArrayList<Tree>();
		min.add(minDup);
		test(geneTree,speciesTree,listLabels,min,speciesNoList,speciesNodesList);
		return min.get(0);
	}
	
	//test all possibility of leaf pruning (maximum 24 species)
	public static void test(Tree geneTree, Tree speciesTree, Map<String,ArrayList<Node>> listLabels, 
			ArrayList<Tree> min,HashMap<String, Integer> speciesNoList,HashMap<String, Node> speciesNodesList) {
		Iterator it = listLabels.entrySet().iterator();
		int l1;
		if(it.hasNext()) {
			l1 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l1 = 1;
		}
		
		int l2;
		if(it.hasNext()) {
			l2 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l2 = 1;
		}
		
		int l3;
		if(it.hasNext()) {
			l3 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l3 = 1;
		}
		
		int l4;
		if(it.hasNext()) {
			l4 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l4 = 1;
		}
		
		int l5;
		if(it.hasNext()) {
			l5 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l5 = 1;
		}
		
		int l6;
		if(it.hasNext()) {
			l6 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l6 = 1;
		}
		
		int l7;
		if(it.hasNext()) {
			l7 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l7 = 1;
		}
		
		int l8;
		if(it.hasNext()) {
			l8 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l8 = 1;
		}
		
		int l9;
		if(it.hasNext()) {
			l9 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l9 = 1;
		}
		
		int l10;
		if(it.hasNext()) {
			l10 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l10 = 1;
		}
		
		int l11;
		if(it.hasNext()) {
			l11 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l11 = 1;
		}
		
		int l12;
		if(it.hasNext()) {
			l12 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l12 = 1;
		}
		
		int l13;
		if(it.hasNext()) {
			l13 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l13 = 1;
		}
		
		int l14;
		if(it.hasNext()) {
			l14 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l14 = 1;
		}
		
		int l15;
		if(it.hasNext()) {
			l15 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l15 = 1;
		}
		
		int l16;
		if(it.hasNext()) {
			l16 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l16 = 1;
		}
		
		int l17;
		if(it.hasNext()) {
			l17 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l17 = 1;
		}
		
		int l18;
		if(it.hasNext()) {
			l18 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l18 = 1;
		}
		
		int l19;
		if(it.hasNext()) {
			l19 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l19 = 1;
		}
		
		int l20;
		if(it.hasNext()) {
			l20 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l20 = 1;
		}
		
		int l21;
		if(it.hasNext()) {
			l21 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l21 = 1;
		}
		
		int l22;
		if(it.hasNext()) {
			l22 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l22 = 1;
		}
		
		int l23;
		if(it.hasNext()) {
			l23 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l23 = 1;
		}
		
		int l24;
		if(it.hasNext()) {
			l24 = ((ArrayList<Node>)((Map.Entry)it.next()).getValue()).size();
		} else {
			l24 = 1;
		}
		
		for(int a = 0; a < l1;a++) {
		for(int b = 0; b < l2;b++) {
		for(int c = 0; c < l3;c++) {
		for(int d = 0; d < l4;d++) {
		for(int e = 0; e < l5;e++) {
		for(int f = 0; f < l6;f++) {
		for(int g = 0; g < l7;g++) {
		for(int h = 0; h < l8;h++) {
		for(int i = 0; i < l9;i++) {
		for(int l = 0; l < l10;l++) {
		for(int k = 0; k < l11;k++) {
		for(int m = 0; m < l12;m++) {
		for(int n = 0; n < l13;n++) {
		for(int o = 0; o < l14;o++) {		
		for(int p = 0; p < l15;p++) {			
		for(int q = 0; q < l16;q++) {
		for(int r = 0; r < l17;r++) {
		for(int s = 0; s < l18;s++) {
		for(int t = 0; t < l19;t++) {
		for(int u = 0; u < l20;u++) {
		for(int v = 0; v < l21;v++) {	
		for(int x = 0; x < l22;x++) {
		for(int y = 0; y < l23;y++) {
		for(int z = 0; z < l24;z++) {	
			
			ArrayList<Integer> newResult = new  ArrayList<Integer>();
			newResult.add(a);
			newResult.add(b);
			newResult.add(c);
			newResult.add(d);
			newResult.add(e);
			newResult.add(f);
			newResult.add(g);
			newResult.add(h);
			newResult.add(i);
			newResult.add(l);
			newResult.add(k);
			newResult.add(m);
			newResult.add(n);
			newResult.add(o);
			newResult.add(p);
			newResult.add(q);
			newResult.add(r);
			newResult.add(s);
			newResult.add(t);
			newResult.add(u);
			newResult.add(v);
			newResult.add(x);
			newResult.add(y);
			newResult.add(z);

			Tree newTree = duplicateTree(geneTree, speciesTree,speciesNoList,speciesNodesList);
			int Counter = 0;
			for (ArrayList<Node> listNode : listLabels.values())  {
	             for(int j = 0; j < listNode.size();j++) {
	            	 if(j != newResult.get(Counter)) {
	            		 int noNode = listNode.get(j).getNoNode();
	            		 pruneLeaf(findNodeGene(newTree,noNode), newTree);
	            		
	            	 }
	             }
				Counter++;
			}
			findCladesSpecies(newTree);
			chooseSpeciesNodeInternalGene(newTree,speciesTree);
			findParent(newTree);
			findDuplicationNodes(newTree,speciesNoList,speciesNodesList);
	   		Tree minDup = min.get(0);
			if(newTree.getNumDuplications() < minDup.getNumDuplications() ) {
					minDup = newTree;
					min.set(0, minDup);
				}
			
			}}}}}}}}}}}}}}}}}}}}}}}}}
	
	
	//Find all leaves labels of the tree
	public static void findLeavesRecursive(Node currentNode,Map<String,ArrayList<Node>> listLabels) {
		if(currentNode != null) {
			findLeavesRecursive(currentNode.getLeftChild(),listLabels);
			findLeavesRecursive(currentNode.getRightChild(),listLabels);
			String label = currentNode.getLabel();
			if(label != null) {
				ArrayList<Node> listNode = listLabels.get(label);
				if(listNode != null ) {
					listNode.add(currentNode);
				} else {
					listNode = new ArrayList<Node>();
					listNode.add(currentNode);
					listLabels.put(label, listNode);
				}
			} 			
		}
	}
	
	
	//Prune all the leaves of the gene tree whose labels are not in the set labels (species tree)
	public static void FiltrergeneTree(Tree geneTree,Set<String> labels  ) {
		Node root = geneTree.getRoot();
		FiltrergeneTreeRecursive(root, labels, geneTree);
	}
	
	public static void FiltrergeneTreeRecursive(Node currentNode,Set<String> labels, Tree geneTree  ) {
		if(currentNode != null) {
			FiltrergeneTreeRecursive(currentNode.getLeftChild(),labels,geneTree);
			FiltrergeneTreeRecursive(currentNode.getRightChild(),labels,geneTree);
			if(currentNode.getRightChild() == null) {
				if(!labels.contains(currentNode.getSpecies())) {
					pruneLeaf(currentNode, geneTree);
				}
			}
		}
	}
	
		//Prune a leaf of the gene tree
		public static void pruneLeaf(Node currentNode, Tree geneTree) {
			Node parent = currentNode.getParent();
				Node neighbour;
				if(parent.getLeftChild() == currentNode) {
					neighbour = parent.getRightChild();
				} else {
					neighbour = parent.getLeftChild();
				}
							
				Node grandParent = parent.getParent();
				if(grandParent != null) {
					if(grandParent.getLeftChild() == parent) {
						grandParent.setLeftChild(neighbour);
					} else {
						grandParent.setRightChild(neighbour);
					}
							
					neighbour.setParent(grandParent);
							
				} else {
					neighbour.setParent(null);
					geneTree.setRoot(neighbour);
				}
		}
	
		//MULPR algorithm
		//A greedy algorithm that tries to prune the leaves in geneTree in a way that fixes the most duplication nodes.
	public static Tree pruningGreedy(Tree geneTree, Tree speciesTree,HashMap<String, Integer> speciesNoList,HashMap<String, Node> speciesNodesList) {
		ArrayList<Tree> listTree = new ArrayList<Tree>();
		Node root = geneTree.getRoot();
		pruningRecursive(root,geneTree,speciesTree,listTree,speciesNoList,speciesNodesList);
		if(listTree.size() != 0) {
			Collections.sort(listTree, new SortByNumDuplications()); 
			return pruningGreedy(listTree.get(0), speciesTree,speciesNoList,speciesNodesList);
		} else {
			return geneTree;
		}
	}
	

	
	//Compute B1, B2, B3 for all node in the subtree of the species tree rooted at speciesLCA
	public static void speciesTreePreprocess(Node duplicationNode, Tree geneTree, Tree speciesTree, Node speciesLCA) {
		HashMap<String, Integer> listeFeuillesGauche = getLeafLabelsMultiplicity (duplicationNode.getLeftChild());
		HashMap<String, Integer> listeFeuillesDroit = getLeafLabelsMultiplicity (duplicationNode.getRightChild());
		HashMap<String, Integer> listeFeuillesComplet = getLeafLabelsMultiplicity (geneTree.getRoot());
		ComputeB1B2B3(speciesLCA, listeFeuillesGauche, listeFeuillesDroit, listeFeuillesComplet);
	}
	
	
	public static  void ComputeB1B2B3 (Node currentNode, HashMap<String, Integer> listeFeuillesGauche, 
											HashMap<String, Integer> listeFeuillesDroit,HashMap<String, Integer> listeFeuillesComplet) {
		if(currentNode != null) {
			ComputeB1B2B3(currentNode.getLeftChild(),listeFeuillesGauche,listeFeuillesDroit,listeFeuillesComplet);
			ComputeB1B2B3(currentNode.getRightChild(),listeFeuillesGauche,listeFeuillesDroit,listeFeuillesComplet);
			currentNode.setB1(false);
			currentNode.setB2(false);
			currentNode.setB3(false);
			if(currentNode.getRightChild() != null) {
				if(currentNode.getRightChild().isB1() && currentNode.getLeftChild().isB1()) {
					currentNode.setB1(true);
				}
				if(currentNode.getRightChild().isB2() && currentNode.getLeftChild().isB2()) {
					currentNode.setB2(true);
				}
				if(currentNode.getRightChild().isB3() && currentNode.getLeftChild().isB3()) {
					currentNode.setB3(true);
				}
			} else {
				
				int geneTreeMultiplicity = 1;
				if(listeFeuillesComplet.get(currentNode.getSpecies()) != null) {
					 geneTreeMultiplicity = listeFeuillesComplet.get(currentNode.getSpecies());
				}
				
				int leftChildMultiplicity = 0;
				if(listeFeuillesGauche.get(currentNode.getSpecies()) != null) {
					leftChildMultiplicity = listeFeuillesGauche.get(currentNode.getSpecies());
				}
				
				int rightChildMultiplicity = 0;
				if(listeFeuillesDroit.get(currentNode.getSpecies()) != null) {
					rightChildMultiplicity = listeFeuillesDroit.get(currentNode.getSpecies());
				}
				
				if((geneTreeMultiplicity - leftChildMultiplicity) > 0){
					currentNode.setB1(true);
				}
				if((geneTreeMultiplicity - rightChildMultiplicity ) > 0){
					currentNode.setB2(true);
				}
				if((geneTreeMultiplicity - leftChildMultiplicity - rightChildMultiplicity ) > 0){
					currentNode.setB3(true);
				}
				
			}
		}
	}
	
	
	//Create a list of all leaf labels with multiplicity of the subtree of the gene tree given in input with the node given in input as root 
	public static HashMap<String, Integer> getLeafLabelsMultiplicity (Node node) {
		HashMap<String, Integer> leafLabelsMultiplicity = new HashMap<String, Integer>();
		getLeavesLabeledMultiplicityRecursive(node,leafLabelsMultiplicity);
		return leafLabelsMultiplicity;
	}
	
	public static void getLeavesLabeledMultiplicityRecursive (Node currentNode, HashMap<String, Integer> leafLabelsMultiplicity) {
		if(currentNode != null) {
			getLeavesLabeledMultiplicityRecursive(currentNode.getLeftChild(),leafLabelsMultiplicity);
			getLeavesLabeledMultiplicityRecursive(currentNode.getRightChild(),leafLabelsMultiplicity);
			if(currentNode.getRightChild() == null) {
				String label = currentNode.getLabel();
				Integer multiplicity = leafLabelsMultiplicity.get(label);
				if(multiplicity != null) {
					leafLabelsMultiplicity.put(label,1 + multiplicity);
				} else {
				leafLabelsMultiplicity.put(label,1);
				}
			}
		}
	}
	
	//Compute the clade of the node given in input
	public static void findClade (Node node) {
		Set<String> clade = new HashSet<String>();
		findCladeRecursive(node,clade);
		node.setCladeLabel(clade);
	}
	
	
	public static void findCladeRecursive(Node currentNode,Set<String> clade) {
		if(currentNode != null) {
			findCladeRecursive(currentNode.getLeftChild(),clade);
			findCladeRecursive(currentNode.getRightChild(),clade);
			if(currentNode.getRightChild() == null) {
				clade.add(currentNode.getLabel());
			} 		
		}
	}
	
	
	public static void pruningRecursive(Node currentNode, Tree geneTree, Tree speciesTree, ArrayList<Tree> listTree,HashMap<String, Integer> speciesNoList,HashMap<String, Node> speciesNodesList) {
		if(currentNode != null) {
			pruningRecursive(currentNode.getLeftChild(),geneTree, speciesTree,listTree,speciesNoList,speciesNodesList);
			pruningRecursive(currentNode.getRightChild(),geneTree, speciesTree,listTree,speciesNoList,speciesNodesList);
			if(currentNode.isDuplication()) {
				int noNodeDuplication = currentNode.getNoNode();
				findClade (currentNode);
				Node speciesLCA = noeudLCA(speciesTree,currentNode.getCladeLabel());
				speciesTreePreprocess(currentNode,geneTree,speciesTree,speciesLCA);
				Set<String> leavesToPrune = new HashSet<String>();
				Tree PrunedTree = eliminerDuplication(currentNode,noNodeDuplication,geneTree,speciesTree,speciesLCA,leavesToPrune,speciesNoList,speciesNodesList);
				if(PrunedTree != null) {
					listTree.add(PrunedTree);
				}
			}
		}
	}	
	

	//Return a pruned tree with the duplication node fixed if possible, and null otherwise
	public static Tree eliminerDuplication(Node currentNode, int noNodeDuplication, Tree geneTree, Tree speciesTree, Node Species
													, Set<String> leavesToPrune,HashMap<String, Integer> speciesNoList,HashMap<String, Node> speciesNodesList) {
		
		Set<String> leavesToPrune1 = Species.getNodesToPrune1();
		Set<String> leavesToPrune2 = Species.getNodesToPrune2();
		
		//Test 4 cases
		Tree resultTree1   = null;
		Tree resultTree2   = null;
		Tree resultTree3   = null;
		Tree resultTree4   = null;
		

		Node speciesLeftChild = Species.getLeftChild();
		Node speciesRightChild = Species.getRightChild();
		
		if(speciesLeftChild == null) {
			speciesLeftChild = Species;
		}
		
		if(speciesRightChild == null) {
			speciesRightChild = Species;
		}
		
		
		//Case no1:
		if(speciesLeftChild.isB2() && speciesRightChild.isB1()) {
			resultTree1   = duplicateTree(geneTree,speciesTree,speciesNoList,speciesNodesList);
			Node duplicationNode1 = findNodeGene(resultTree1,noNodeDuplication);
			
			leavesToPrune1.addAll(leavesToPrune);
			leavesToPrune2.addAll(leavesToPrune);
			
			pruneLeavesRecursive(duplicationNode1.getRightChild(), leavesToPrune2, resultTree1);
			pruneLeavesRecursive(duplicationNode1.getLeftChild(), leavesToPrune1, resultTree1);
			
			//findCladeslabel(resultTree1);

			findCladesSpecies(resultTree1);
			chooseSpeciesNodeInternalGene(resultTree1,speciesTree);
			findParent(resultTree1);
			findDuplicationNodes(resultTree1,speciesNoList,speciesNodesList);
			
			findNoNodeGene(resultTree1);
		}
		
		//Case no2:
				
		if(speciesRightChild.isB2() && speciesLeftChild.isB1()) {
			resultTree2   = duplicateTree(geneTree,speciesTree,speciesNoList,speciesNodesList);
			Node duplicationNode2 = findNodeGene(resultTree2,noNodeDuplication);
			
			leavesToPrune1.addAll(leavesToPrune);
			leavesToPrune2.addAll(leavesToPrune);
			
			pruneLeavesRecursive(duplicationNode2.getRightChild(), leavesToPrune1, resultTree2);
			pruneLeavesRecursive(duplicationNode2.getLeftChild(), leavesToPrune2, resultTree2);

			//findCladeslabel(resultTree2);

			findCladesSpecies(resultTree2);
			chooseSpeciesNodeInternalGene(resultTree2,speciesTree);
			findParent(resultTree2);
			findDuplicationNodes(resultTree2,speciesNoList,speciesNodesList);
			
			findNoNodeGene(resultTree2);
		}
		
		//Case no3:			
		
		if(Species.isB2()) {
			resultTree3   = duplicateTree(geneTree,speciesTree,speciesNoList,speciesNodesList);
			Node duplicationNode3 = findNodeGene(resultTree3,noNodeDuplication);
		//	System.out.println("cas no3");
						
			pruneLeavesRecursive(duplicationNode3.getRightChild(), leavesToPrune1, resultTree3);
			pruneLeavesRecursive(duplicationNode3.getRightChild(), leavesToPrune2, resultTree3);
		
		//	findCladeslabel(resultTree3);

			findCladesSpecies(resultTree3);
			chooseSpeciesNodeInternalGene(resultTree3,speciesTree);
			findParent(resultTree3);
			findDuplicationNodes(resultTree3,speciesNoList,speciesNodesList);
			
			findNoNodeGene(resultTree3);
		}
		
		//Case no4:
				
		if(Species.isB1()) {
			resultTree4   = duplicateTree(geneTree,speciesTree,speciesNoList,speciesNodesList);
			Node duplicationNode4 = findNodeGene(resultTree4,noNodeDuplication);

			pruneLeavesRecursive(duplicationNode4.getLeftChild(), leavesToPrune1, resultTree4);
			pruneLeavesRecursive(duplicationNode4.getLeftChild(), leavesToPrune2, resultTree4);

			findCladesSpecies(resultTree4);
			chooseSpeciesNodeInternalGene(resultTree4,speciesTree);
			findParent(resultTree4);
			findDuplicationNodes(resultTree4,speciesNoList,speciesNodesList);
			
			findNoNodeGene(resultTree4);
		}
		
		
		//If the 4 cases are impossible
		if(resultTree1 == null && resultTree2 == null && resultTree3 == null && resultTree4 == null ) {
			Tree resultTree5   = null;
		if(speciesRightChild.isB3()) {
			if(!leavesToPrune1.isEmpty()) {
	
				leavesToPrune.addAll(leavesToPrune1);
				
				//Find new lCA
				Node Species5 = Species.getLeftChild();
				while(true) {
					if(Species5.getRightChild() != null) {
						if (Species5.getRightChild().getNumLeaves() == 0) {
							Species5 = Species5.getLeftChild();
						} else if(Species5.getLeftChild().getNumLeaves() == 0) {
							Species5 = Species5.getRightChild();
						} else {
							break;
						}
					} else {
						break;
					}	
				}
				
					resultTree5 = eliminerDuplication(currentNode,noNodeDuplication, geneTree, speciesTree,Species5,leavesToPrune,speciesNoList,speciesNodesList); 
				
			} 
		} 
			if(resultTree5 != null) {
				return resultTree5;
			} else {
				Tree resultTree6   = null;
				if(speciesLeftChild.isB3()) {
					if(!leavesToPrune2.isEmpty()) {

						leavesToPrune.addAll(leavesToPrune2);
						
						//Find new lCA
						Node Species6 = Species.getRightChild();
						while(true) {
							if(Species6.getRightChild() != null) {
								if (Species6.getRightChild().getNumLeaves() == 0) {
									Species6 = Species6.getLeftChild();
								} else if(Species6.getLeftChild().getNumLeaves() == 0) {
									Species6 = Species6.getRightChild();
								} else {
									break;
								}
							} else {
								break;
							}	
						}

							resultTree6 = eliminerDuplication(currentNode,noNodeDuplication, geneTree, speciesTree,Species6,leavesToPrune,speciesNoList,speciesNodesList);
					} 
				}
				return resultTree6;
			}
		} else {
			ArrayList<Tree> listTree = new ArrayList<Tree>();
			if(resultTree1 != null) {
				listTree.add(resultTree1);
			}
			if(resultTree2 != null) {
				listTree.add(resultTree2);
			}
			if(resultTree3 != null) {
				listTree.add(resultTree3);
			}
			if(resultTree4 != null) {
				listTree.add(resultTree4);
			}
			Collections.sort(listTree, new SortByNumDuplications()); 
			return listTree.get(0);
		}
	}
	
	
	//In a gene tree, return the node whose number is noNode
	public static Node findNodeGene(Tree geneTree,int noNode) {
		
		Node root = geneTree.getRoot();		
		return findNodeGeneRecursive(root,noNode);
	}
	
	//In a gene tree, return the node whose number is noNode
	public static Node findNodeGeneRecursive(Node currentNode,int noNode) {
		if(noNode > currentNode.getNoNode()) {
			return findNodeGeneRecursive(currentNode.getRightChild(),noNode);
		} else if (noNode < currentNode.getNoNode()) {
			return findNodeGeneRecursive(currentNode.getLeftChild(),noNode);
		} else {
			return currentNode;
		}
	}
	
	//Give number to the nodes of a gene tree (infix order)
	public static void findNoNodeGene(Tree geneTree) {
		
		geneTree.setCounterNodes(0);
		Node root = geneTree.getRoot();		
		findNoNodeGeneRecursive(root,geneTree );
	}
	
	//Give number to the nodes of a gene tree (infix order)
	public static void findNoNodeGeneRecursive(Node currentNode,Tree geneTree) {

		if(currentNode != null) {
			findNoNodeGeneRecursive(currentNode.getLeftChild(),geneTree);
			int Counter = geneTree.getCounterNodes();
			currentNode.setNoNode(Counter);
			geneTree.setCounterNodes(Counter + 1);
			findNoNodeGeneRecursive(currentNode.getRightChild(),geneTree);
			
		}
	}
	
	//Duplicate a gene tree
	public static Tree duplicateTree(Tree geneTree,Tree speciesTree, HashMap<String, Integer> speciesNoList, HashMap<String, Node> speciesNodesList) {
		String newickSpecies = geneTree.getRoot().toString3();
		String newickGene = geneTree.getRoot().toString4();
		
		Tree tree = creergeneTree(newickSpecies,newickGene);
		findCladesSpecies(tree);
		chooseSpeciesNodeInternalGene(tree,speciesTree);
		findParent(tree);
		findDuplicationNodes(tree,speciesNoList,speciesNodesList);
		findNoNodeGene(tree);
		return tree;
	}
	
	//Find the leaves to prune to get lower in the species tree 
	public static void findLeavesToPrune(Tree speciesTree) {
		
		Node root = speciesTree.getRoot();		
		findLeavesToPruneRecursive(root);
	}
			
	//Find the leaves to prune to get lower in the species tree 
	public static void findLeavesToPruneRecursive(Node currentNode) {

		if(currentNode != null) {
			findLeavesToPruneRecursive(currentNode.getLeftChild());
			findLeavesToPruneRecursive(currentNode.getRightChild());
			if(currentNode.getRightChild() != null) {
				Set<String> clade = new HashSet<String>(); 
				clade.addAll(currentNode.getCladeSpecies());
				clade.removeAll(currentNode.getLeftChild().getCladeSpecies());
				// ou clade.addAll(currentNode.getRightChild().getCladeSpecies()); ?
				
				Set<String> clade2 = new HashSet<String>(); 
				clade2.addAll(currentNode.getCladeSpecies());
				clade2.removeAll(currentNode.getRightChild().getCladeSpecies());
				// ou clade.addAll(currentNode.getLeftChild().getCladeSpecies()); ?
				
				currentNode.setNodesToPrune1(clade);
				currentNode.setNodesToPrune2(clade2);
			} else {
				Set<String> clade = new HashSet<String>(); 
				clade.addAll(currentNode.getCladeSpecies());
				currentNode.setNodesToPrune1(clade);
				Set<String> clade2 = new HashSet<String>(); 
				currentNode.setNodesToPrune2(clade2);   //empty
			}
		}
	}
	
	
	//Prune all leaves that are descendant of currentNode (or currentNode itself) whose labels are in listSpecies
	public static void pruneLeavesRecursive(Node currentNode, Set<String> listSpecies, Tree geneTree) {

		if(currentNode != null) {
			pruneLeavesRecursive(currentNode.getLeftChild(),listSpecies,geneTree);
			pruneLeavesRecursive(currentNode.getRightChild(),listSpecies,geneTree);
			if(currentNode.getRightChild() == null) {
				if(listSpecies.contains(currentNode.getSpecies())) {
					Node parent = currentNode.getParent();
					Node neighbour;
					if(parent.getLeftChild() == currentNode) {
						neighbour = parent.getRightChild();
					} else {
						neighbour = parent.getLeftChild();
					}
					
					Node grandParent = parent.getParent();
					if(grandParent != null) {
						if(grandParent.getLeftChild() == parent) {
							grandParent.setLeftChild(neighbour);
						} else {
							grandParent.setRightChild(neighbour);
						}
					
						neighbour.setParent(grandParent);
					
					} else {
						neighbour.setParent(null);
						geneTree.setRoot(neighbour);
					}
				}
			}	
		}
	}	
	
	
	//Find the parents of all nodes of the tree given in input
	public static void findParent(Tree geneTree) {
		
		Node root = geneTree.getRoot();		
		findParentRecursive(root,null);
	}
	public static void findParentRecursive(Node currentNode, Node parentNode) {

		if(currentNode != null) {
			findParentRecursive(currentNode.getLeftChild(),currentNode);
			findParentRecursive(currentNode.getRightChild(),currentNode);
			
			currentNode.setParent(parentNode);	
		}
	}
		
	//Preprocess the species tree to find the duplication nodes
	public static void PreProcessS(Tree speciesTree) {
		
		speciesTree.setCounterNodes(1);
		Node root = speciesTree.getRoot();		
		PreProcessSRecursive(root,speciesTree);
	}
	
	//Preprocess the species tree to find the duplication nodes
	public static void PreProcessSRecursive (Node currentNode,Tree speciesTree) {
		if(currentNode != null) {
			int Counter = speciesTree.getCounterNodes();
			currentNode.setNoNode(Counter);
			speciesTree.setCounterNodes(Counter + 1);
			PreProcessSRecursive(currentNode.getLeftChild(),speciesTree);
			PreProcessSRecursive(currentNode.getRightChild(),speciesTree);
			
		}
	}
	public static void computeSpeciesNoList(Node currentNode, HashMap<String, Integer> speciesNoList) {
		if(currentNode != null) {
			computeSpeciesNoList(currentNode.getLeftChild(), speciesNoList);
			computeSpeciesNoList(currentNode.getRightChild(), speciesNoList);
			if(currentNode.getRightChild() == null) {
				speciesNoList.put(currentNode.getSpecies(), currentNode.getNoNode());
			}
		}
	}
	
	public static void computeSpeciesNodesList(Node currentNode, HashMap<String, Node> speciesNodesList) {
		if(currentNode != null) {
			computeSpeciesNodesList(currentNode.getLeftChild(), speciesNodesList);
			computeSpeciesNodesList(currentNode.getRightChild(), speciesNodesList);
			if(currentNode.getRightChild() == null) {
				speciesNodesList.put(currentNode.getSpecies(), currentNode);
			}
		}
	}
	
	//Find the duplication nodes (Algorithm from Christian M. Zmasek and Sean R. Eddy (A simple algorithm to infer gene duplication and speciation events on a gene tree) )	
	// This algorithm run in O(n^2), but has good performance with real data.
	public static void findDuplicationNodes(Tree geneTree, HashMap<String, Integer> speciesNoList, HashMap<String, Node> speciesNodesList) {
		Node root = geneTree.getRoot();		
		geneTree.setNumDuplications(0);
		findDuplicationNodesRecursive(root,geneTree,speciesNoList,speciesNodesList);
	}
	
	public static void findDuplicationNodesRecursive(Node currentNode,Tree geneTree, HashMap<String, Integer> speciesNoList, HashMap<String,Node> speciesNodesList ) {
		if(currentNode != null) {
			findDuplicationNodesRecursive(currentNode.getLeftChild(),geneTree, speciesNoList,  speciesNodesList);
			findDuplicationNodesRecursive(currentNode.getRightChild(),geneTree, speciesNoList,  speciesNodesList);
			if(currentNode.getRightChild() != null) {
				int Mg1 = currentNode.getLeftChild().getMg();
				int Mg2 = currentNode.getRightChild().getMg();
				Node species1 = currentNode.getLeftChild().getSpeciesLCA();
				Node species2 = currentNode.getRightChild().getSpeciesLCA();
				while (Mg1 != Mg2) {
					if(Mg1 > Mg2) {
						species1 = species1.getParent();
						Mg1 = species1.getNoNode();
					} else {
						species2 = species2.getParent();
						Mg2 = species2.getNoNode();
					}
				}
				currentNode.setMg(Mg1);
				currentNode.setSpeciesLCA(species1);
				
				if((currentNode.getMg() == currentNode.getLeftChild().getMg() ) 
								|| (currentNode.getMg() == currentNode.getRightChild().getMg() ) ) {
					currentNode.setDuplication(true);
					geneTree.setNumDuplications(geneTree.getNumDuplications() + 1);
				} else {
					currentNode.setDuplication(false);			
				}			
			} else {
				currentNode.setDuplication(false);
				currentNode.setMg(speciesNoList.get(currentNode.getSpecies()));
				currentNode.setSpeciesLCA(speciesNodesList.get(currentNode.getSpecies()));
			}
		}
	}

	//Give a species name (LCA mapping) to the internal nodes of the genes tree
	public static void chooseSpeciesNodeInternalGene(Tree geneTree,Tree speciesTree) {
		Node root = geneTree.getRoot();		
		chooseSpeciesNodeInternalGeneRecursive(root,speciesTree);
	}
		
	//Give a species name (LCA mapping) to the internal nodes of the genes tree
	public static void chooseSpeciesNodeInternalGeneRecursive(Node currentNode,Tree speciesTree ) {
		if(currentNode != null) {
			chooseSpeciesNodeInternalGeneRecursive(currentNode.getLeftChild(),speciesTree);
			chooseSpeciesNodeInternalGeneRecursive(currentNode.getRightChild(),speciesTree);
			if(currentNode.getRightChild() != null) {
				String Species = SpeciesLCA(speciesTree,currentNode.getCladeSpecies());
				currentNode.setSpecies(Species);
			}	
		}
	}
	
	
	//Find all clades of the tree
	public static void findCladesSpecies(Tree tree) {
		Node root = tree.getRoot();		
		findCladesRecursiveSpecies(root);
	}
	
	//Find all clades of the tree
	public static void findCladesRecursiveSpecies(Node currentNode) {
		if(currentNode != null) {
			findCladesRecursiveSpecies(currentNode.getLeftChild());
			findCladesRecursiveSpecies(currentNode.getRightChild());
			Set<String> clade = new HashSet<String>(); 
			if(currentNode.getRightChild() != null) {
				clade.addAll(currentNode.getLeftChild().getCladeSpecies());
				clade.addAll(currentNode.getRightChild().getCladeSpecies());
				currentNode.setCladeSpecies(clade);
			} else {
				clade.add(currentNode.getSpecies());
				currentNode.setCladeSpecies(clade);
			}
			
		}
	}
	
	//Return the node of the species tree corresponding to the given clade (1)
	public static Node noeudLCA(Tree tree,Set<String> clade) {
		Node root = tree.getRoot();	
		SpeciesLCARecursive1(root,clade);
		Node LCA = SpeciesLCARecursive3(root,clade);
		return LCA;
	}	
	
	
	//Return the species of the species tree corresponding to the given clade 
	public static String SpeciesLCA(Tree tree,Set<String> clade) {
		Node root = tree.getRoot();	
		SpeciesLCARecursive1(root,clade);
		Node LCA = SpeciesLCARecursive3(root,clade);
		return LCA.getSpecies();
	}	
	
	
	public static void SpeciesLCARecursive1(Node currentNode,Set<String> clade) {
		if(currentNode != null) {
			SpeciesLCARecursive1(currentNode.getLeftChild(),clade);
			SpeciesLCARecursive1(currentNode.getRightChild(),clade);
			if(currentNode.getRightChild() != null) {
				currentNode.setNumLeaves(currentNode.getLeftChild().getNumLeaves() + currentNode.getRightChild().getNumLeaves());
			} else {
				if(clade.contains(currentNode.getSpecies())) {
					currentNode.setNumLeaves(1);
				} else {
					currentNode.setNumLeaves(0);
				}
			}	
		}
	}
	

	public static Node SpeciesLCARecursive3(Node currentNode,Set<String> clade) {
		if(currentNode.getLeftChild() != null && (currentNode.getLeftChild().getNumLeaves() == clade.size())) {
			return SpeciesLCARecursive3(currentNode.getLeftChild(),clade);
		} else if (currentNode.getRightChild() != null && (currentNode.getRightChild().getNumLeaves() == clade.size())) {
			return SpeciesLCARecursive3(currentNode.getRightChild(),clade);
		} else {
			return currentNode;
		}
	}
	
	//Give a species name to the internal nodes of the species tree
	public static void chooseSpeciesNodeInternalSpecies(Tree tree) {
		Node root = tree.getRoot();		
		chooseSpeciesNodeInternalSpeciesRecursive(root,tree);
	}
	
	//Give a species name to the internal nodes of the species tree
	public static void chooseSpeciesNodeInternalSpeciesRecursive(Node currentNode,Tree tree) {
		if(currentNode != null) {
			chooseSpeciesNodeInternalSpeciesRecursive(currentNode.getLeftChild(),tree);
			chooseSpeciesNodeInternalSpeciesRecursive(currentNode.getRightChild(),tree);
			if(currentNode.getRightChild() != null) {
				int Counter = tree.getCounterSpecies();
				currentNode.setSpecies(Integer.toString(Counter));
				tree.setCounterSpecies(Counter + 1);
			}	
		}
	}
	
	//Create the species tree 
	public static Tree creerspeciesTree(String speciesNewick) {
		Node root = new Node();
			
		//Remove white spaces in the string
		speciesNewick = speciesNewick.replaceAll("\\s+","");
			
		//Remove the ";" at the end of the strings
		speciesNewick = speciesNewick.replaceAll(";","");
			
		getspeciesTree(root,speciesNewick);
			
		Tree tree = new Tree(root);
			
		return tree;	
	}
	
	//Create the gene tree
	public static Tree creergeneTree(String speciesNewick,String labelsNewick ) {
		Node root = new Node();
			
		//Remove white spaces in the strings
		speciesNewick = speciesNewick.replaceAll("\\s+","");
		labelsNewick = labelsNewick.replaceAll("\\s+","");
		
		//Remove the ";" at the end of the strings
		speciesNewick = speciesNewick.replaceAll(";","");
		labelsNewick = labelsNewick.replaceAll(";","");
		
		getgeneTree(root,speciesNewick,labelsNewick);
			
		Tree tree = new Tree(root);
			
		return tree;	
	}
	
	//Put in the node root given in input the gene tree corresponding to the tree given in input
	public static void getgeneTree(Node root,String speciesNewick,String labelsNewick) {
		speciesNewick = speciesNewick.substring(1, speciesNewick.length() -1 );
		labelsNewick = labelsNewick.substring(1, labelsNewick.length() -1 );
			
		if(speciesNewick.substring(0, 1).equals("(")) {
			//Left child
			
			//Species
			int CounterParentheses = 1;
			int i = 1;
			for(i = 1; i < speciesNewick.length(); i++) {
				if(speciesNewick.substring(i, i+1).equals("(")) {
					CounterParentheses++;
				} else if(speciesNewick.substring(i, i+1).equals(")")) {
					CounterParentheses--;
				}
						
				if(CounterParentheses == 0) {
						break;
				}
			}
			
			//Label
			CounterParentheses = 1;
			int j = 1;
			for(j = 1; j < labelsNewick.length(); j++) {
				if(labelsNewick.substring(j, j+1).equals("(")) {
					CounterParentheses++;
				}	else if(labelsNewick.substring(j, j+1).equals(")")) {
					CounterParentheses--;
				}
				
				if(CounterParentheses == 0) {
					break;
				}
			}
			
			
			Node leftChild = new Node();
			root.setLeftChild(leftChild);
			getgeneTree(leftChild,speciesNewick.substring(0,i+1),labelsNewick.substring(0,j+1));
				
			//Right child
			i = i+2;
			j = j+2;
			if(speciesNewick.substring(i,i + 1).equals("(")) {

				Node rightChild = new Node();
				root.setRightChild(rightChild);
				getgeneTree(rightChild,speciesNewick.substring(i,speciesNewick.length()),labelsNewick.substring(j,labelsNewick.length()));
			} else {
				Node rightChild = new Node();
				rightChild.setSpecies(speciesNewick.substring(i,speciesNewick.length()));
				rightChild.setLabel(labelsNewick.substring(j,labelsNewick.length()));
				root.setRightChild(rightChild);

			}	
		} else {
			//Left child
			
			//Species
			int i  = 0;
			for(i = 0; i < speciesNewick.length(); i++) {
				if(speciesNewick.substring(i, i+1).equals(",")) {
					break;
				}
			}
			
			//Label
			int j = 0;
			for(j = 0; j < labelsNewick.length(); j++) {
				if(labelsNewick.substring(j, j+1).equals(",")) {
					break;
				}
			}
			
			Node leftChild = new Node();
			leftChild.setSpecies(speciesNewick.substring(0,i));
			leftChild.setLabel(labelsNewick.substring(0,j));
			root.setLeftChild(leftChild);
				
			//Right child
			i = i+1;
			j = j+1;
			if(speciesNewick.substring(i,i + 1).equals("(")) {

				Node rightChild = new Node();
				root.setRightChild(rightChild);
				getgeneTree(rightChild,speciesNewick.substring(i,speciesNewick.length()),labelsNewick.substring(j,labelsNewick.length()) );
			} else {
				Node rightChild = new Node();
				rightChild.setSpecies(speciesNewick.substring(i,speciesNewick.length()));
				rightChild.setLabel(labelsNewick.substring(j,labelsNewick.length()));
				root.setRightChild(rightChild);
			}
		}
	}	
	
	//Put in the node root given in input the species tree corresponding to the tree given in input
	public static void getspeciesTree(Node root,String speciesNewick) {
		speciesNewick = speciesNewick.substring(1, speciesNewick.length() -1 );
			
		if(speciesNewick.substring(0, 1).equals("(")) {
			//Left child
			int CounterParentheses = 1;
			int i = 1;
			for(i = 1; i < speciesNewick.length(); i++) {
				if(speciesNewick.substring(i, i+1).equals("(")) {
					CounterParentheses++;
				} else if(speciesNewick.substring(i, i+1).equals(")")) {
					CounterParentheses--;
				}
						
				if(CounterParentheses == 0) {
						break;
				}
			}
			Node leftChild = new Node();
			root.setLeftChild(leftChild);
			getspeciesTree(leftChild,speciesNewick.substring(0,i+1));
				
			//Right child
			i = i+2;
			if(speciesNewick.substring(i,i + 1).equals("(")) {

				Node rightChild = new Node();
				root.setRightChild(rightChild);
				getspeciesTree(rightChild,speciesNewick.substring(i,speciesNewick.length()));
			} else {
				Node rightChild = new Node();
				rightChild.setSpecies(speciesNewick.substring(i,speciesNewick.length()));
				root.setRightChild(rightChild);

			}	
		} else {
			//Left child
			int i  = 0;
			for(i = 0; i < speciesNewick.length(); i++) {
				if(speciesNewick.substring(i, i+1).equals(",")) {
					break;
				}
			}
			Node leftChild = new Node();
			leftChild.setSpecies(speciesNewick.substring(0,i));
			root.setLeftChild(leftChild);
				
			//Right child
			i = i+1;
			if(speciesNewick.substring(i,i + 1).equals("(")) {

				Node rightChild = new Node();
				root.setRightChild(rightChild);
				getspeciesTree(rightChild,speciesNewick.substring(i,speciesNewick.length()));
			} else {
				Node rightChild = new Node();
				rightChild.setSpecies(speciesNewick.substring(i,speciesNewick.length()));
				root.setRightChild(rightChild);
			}
		}
	}
}
